/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <pcap.h>
#include <dlfcn.h>
#include <signal.h>
#include <stdlib.h>
#ifdef GETOPT_H
#include <getopt.h>
#endif
#ifdef UNISTD_H
#include <unistd.h>
#endif
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include "passifist.h"
#include "store.h"
#include "plugin.h"
#include "errors.h"

struct plugin_t *loadplugin(char *pName) {

  struct plugin_t *pNewPlugin;
  char *error;

  pNewPlugin = (struct plugin_t *) malloc( sizeof( struct plugin_t ) );

  memset( pNewPlugin->name, 0, sizeof( pNewPlugin->name ) );
  strncpy( pNewPlugin->name, pName, sizeof( pNewPlugin->name ) -1 );
  pNewPlugin->name[sizeof( pNewPlugin->name )] = 0;

  if ( !pNewPlugin ) {
    fprintf(stderr, "FAILED to allocate memory for plugins...\n");
    return NULL;
  }

  pNewPlugin->pHandle = dlopen(pName, RTLD_LAZY);
  
  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_stats = dlsym(pNewPlugin->pHandle, "plugin_stats");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_manglepacket = dlsym(pNewPlugin->pHandle, "plugin_manglepacket");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_stop = dlsym(pNewPlugin->pHandle, "plugin_stop");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_packetcount = dlsym(pNewPlugin->pHandle, "plugin_packetcount");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_getfilter = dlsym(pNewPlugin->pHandle, "plugin_getfilter");
  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_gethost = dlsym(pNewPlugin->pHandle, "plugin_gethost");
  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_freehost = dlsym(pNewPlugin->pHandle, "plugin_freehost");
  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  pNewPlugin->plugin_getdesc = dlsym( pNewPlugin->pHandle, "plugin_getdesc");
  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return NULL;
  }

  /*
  pNewPlugin->next = pPlugins;
  pPlugins = pNewPlugin;
  */

  return pNewPlugin;
}

struct plugin_t *addplugin( struct plugin_t *pPlugins, struct plugin_t *pNew ) {
  pNew->next = pPlugins;
  pPlugins = pNew;

  return pPlugins;

}

void killthread() {
  pthread_cancel( plugth );
}

void *ctrlbreakhandler(void *tmp) {

  sigset_t signalSet = {{0}};
  int signal = 0;

  sigemptyset (&signalSet);
  sigaddset (&signalSet, SIGINT);
  sigprocmask (SIG_BLOCK, &signalSet, NULL);
  
  sigwait(&signalSet, &signal);
  killthread();
  return NULL;
}

void initializectrlbreakhandler() {

  sigset_t  signalSet = {{0}};
  struct sigaction action;

  /* Reset the action struct */
  memset(&action, 0x0, sizeof(action));

  /* If Ctrl+C/Ctrl+Break Signal is received, the threads are to 
   * destroyed and prompt should appear.*/
  sigfillset (&signalSet);
  sigdelset (&signalSet, SIGINT);
  sigprocmask (SIG_BLOCK, &signalSet, NULL);
  sigfillset (&action.sa_mask);

}

void usage( char **argv ) {

  fprintf(stderr, "%s: -i|-r -d [options]\n", argv[0] );
  fprintf(stderr, "\t-r <pcapfile>\n");
  fprintf(stderr, "\t-i <iface>\n");
  fprintf(stderr, "\t-S list storageplugins\n");
  fprintf(stderr, "\t-P list protoplugins\n");
  fprintf(stderr, "\t-U connection url\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "e.g. %s -r /tmp/foo.dmp -U \"provider=TXT;filename=foo\"\n",
	  argv[0]);

}

void callback(u_char *useless, const struct pcap_pkthdr *pkthdr, 
	      const u_char *pPacket) {

  struct plugin_t *pPlug;
  struct host *pHost;
  int nErr;
  
  /* we do not want to get cancelled in here :) */
  pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

  pPlug = pPlugins;

  while ( pPlug ) {

    if ( pPlug->plugin_manglepacket( (u_char *)pPacket, 
				     pkthdr->len, nVerbose ) == 0 ) {

      pHost = pPlug->plugin_gethost();

      if ( pHost ) {
	if ( ( nErr = store_storehost( pHost )  ) != 0 ) {

	  if ( nErr == STORE_STOREHOST_FAILED )
	    fprintf(stderr, "WARNING: Failed to store host\n");
	  else if ( nErr == STORE_STOREPROP_FAILED )
	    fprintf(stderr, "WARNING: Failed to store property\n");
	  else if ( nErr == STORE_CREATECHECKSUM_FAILED )
	    fprintf(stderr, "WARNING: Failed to create md5 checksum\n");
	  /* else if ( nErr == STORE_STOREHOST_DUPLICATE )
	     fprintf(stderr, "INFO: Tried to save duplicate host\n");
	  */
	}

	pPlug->plugin_freehost();
      }
      
      break;
    }

    pPlug = pPlug->next;
  }

  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

}

void *startsniffer(struct sniff_t *pSniff) {
  
  int nErr = 0;

  //pcap_loop( (pcap_t *)pDescr, -1, (void *)callback, NULL );

  while ( 1 ) {
    nErr = pcap_dispatch( (pcap_t *)pSniff->pDescr, 1, (void *)callback, NULL );
    if ( nErr == 0 && pSniff->mode == SNIFFMODE_FILE )
      break;

    pthread_testcancel();
  }

  return NULL;
}

void displaystats( ) {

  struct plugin_t *pPlug;
  u_int32_t nPackets = 0;

  pPlug = pPlugins;

  while ( pPlug ) {
    nPackets += pPlug->plugin_packetcount();
    pPlug = pPlug->next;
  }

  fprintf(stderr, "\rpackets: %d\n", nPackets);
  nPackets = 0;
  
}

void liststorageplugins() {

  DIR *pDir;
  struct dirent *pEntry;
  char pluginname[50];
  char *pDesc;

  pDir = opendir("./db");

  if ( !pDir ) {
    fprintf(stderr, "Failed to open db directory\n");
    return;
  }

  fprintf(stdout, "Storage plugins:\n");

  while ( ( pEntry = readdir( pDir ) ) ) {

    if ( strcmp(pEntry->d_name, "..") != 0 &&
	 strcmp(pEntry->d_name, "." ) != 0 ) {

      memset( pluginname, 0, sizeof(pluginname) );
      strncpy( pluginname, "db/", 3 );
      strncat( pluginname, pEntry->d_name, 20 );

      pDesc = store_getplugdesc( pluginname );
      fprintf(stderr, "\t%s\n", pDesc );
      free( pDesc );
      
    }

  }

  closedir( pDir );

}

void listprotoplugins() {

  DIR *pDir;
  struct dirent *pEntry;
  struct plugin_t *pPlug;
  char pluginname[50];
  char *pDesc;

  pDir = opendir("./plugins");

  if ( !pDir ) {
    fprintf(stderr, "Failed to open plugins directory\n");
    return;
  }

  fprintf(stdout, "Proto plugins:\n");

  while ( ( pEntry = readdir( pDir ) ) ) {

    if ( strcmp(pEntry->d_name, "..") != 0 &&
	 strcmp(pEntry->d_name, "." ) != 0 ) {

      memset( pluginname, 0, sizeof(pluginname) );
      strncpy( pluginname, "plugins/", 8 );
      strncat( pluginname, pEntry->d_name, 20 );

      pPlug = loadplugin( pluginname );
      
      if ( pPlug ) {
	pDesc = pPlug->plugin_getdesc();
	fprintf(stderr, "\t%s\n", pDesc );
	dlclose( pPlug->pHandle );
      }

    }

  }

  closedir( pDir );

}

int main( int argc, char **argv ) {

  char pcaperr[PCAP_ERRBUF_SIZE];
  struct bpf_program bpf;

  pthread_t wthread;

  struct plugin_t *pPlug;
  struct dirent *pEntry;
  struct pcap_stat ps;
  struct sniff_t sniff;

  int c, nLen, nOffset, nErr, nRecordCount = 0;
  char *iface = NULL, *infile = NULL;
  char pluginname[50];
  char pFilter[1024];
  DIR *pDir;
  int nPluginCount = 0;
  char connURL[1024];

  bzero( connURL, sizeof( connURL ) );
  fprintf( stdout, "%s", pVer );

  while (1) {

    c = getopt(argc, argv, "i:r:v:U:SPs:h");

    if ( c == -1 ) 
      break;

    switch (c) {

    case 'h': /* print usage */
      usage(argv);
      exit(1);
      break;
    case 'i': /* specify sniffer interface */
      if ( getuid() != 0 ) {
	fprintf( stderr, "ERROR: You need to be root to go promisc\n");
	exit(1);
      }
      iface = optarg;
      sniff.mode = SNIFFMODE_LIVE;
      break;
    case 'r': /* read dump file */
      infile = optarg;
      sniff.mode = SNIFFMODE_FILE;
      break;
    case 'U':
      strncpy( connURL, optarg, sizeof( connURL ) -1 );
      break;
    case 'v':
      nVerbose = 1;
      break;
    case 'S': /* list storage plugins */
      liststorageplugins();
      exit(1);
      break;
    case 'P':
      listprotoplugins();
      exit(1);
      break;
    default:
      usage(argv);
      exit(1);
    }

  }
  
  if ( iface == NULL && infile == NULL ) {
    usage(argv);
    exit(1);
  }

  pDir = opendir("./plugins");

  if ( !pDir ) {
    fprintf(stderr, "Failed to open plugins directory\n");
    exit(1);
  }

  while ( ( pEntry = readdir( pDir ) ) ) {
    if ( strcmp(pEntry->d_name, "..") != 0 &&
	 strcmp(pEntry->d_name, "." ) != 0 ) {

      memset( pluginname, 0, sizeof(pluginname) );
      strncpy( pluginname, "plugins/", sizeof( pluginname ) );
      strncat( pluginname, pEntry->d_name, 20 );

      if ( ( pPlug = loadplugin( pluginname ) ) == NULL ) {
	fprintf(stderr, "Failed to load plugin %s\n", pluginname);
      }
      else {
	pPlugins = addplugin( pPlugins, pPlug );
	nPluginCount ++;
      }

    }

  }

  closedir( pDir );

  if ( nPluginCount == 0 )
    exit(1);

  if ( iface ) {

    initializectrlbreakhandler();
    
    if (pthread_create (&wthread, NULL, (void*)ctrlbreakhandler, NULL) != 0) {
      printf ("Handler thread creation failed\n");
      return 0;
    }

  }

  fprintf(stdout, "[x] Connecting to provider ...\n");

  if ( ( nErr = store_connect( connURL ) ) != 0 ) {

    if ( nErr == STORE_PROVIDER_RESOLVE_FAILED ) {
      fprintf(stderr, "Could not resolve provider, ");
      fprintf(stderr, "Try listing available storage providers using -S\n");
    }
    else {
      fprintf(stderr, "Connection failed! Make sure you enclose the connection URL in quotes\n");
    exit(1);
    }

  }
  fprintf(stdout, "[x] Connected to provider\n");

  pPlug = pPlugins;
  nOffset = 0;

  while ( pPlug ) {
    /* plug filter + 2 paranthesis + or statement */
    nLen = strlen(pPlug->plugin_getfilter()) + 2 + 4;

    if ( nLen > sizeof( pFilter ) ) {
      fprintf(stderr, "Filter to long\n");
      exit(1);
    }

    strncat(pFilter, "(", sizeof( pFilter ) - strlen( pFilter ) );
    strncat(pFilter, pPlug->plugin_getfilter(), sizeof( pFilter ) - strlen( pFilter) );
    strncat(pFilter, ")", sizeof( pFilter ) - strlen( pFilter ) );

    if ( pPlug->next )
      strncat( pFilter, " or ", sizeof( pFilter ) - strlen( pFilter ) );

    pPlug = pPlug->next;
  }

  //fprintf(stderr, "[x] Setting filter %s\n", pFilter);

  if ( iface != NULL ) {
    fprintf( stderr, "[x] Live on interface=%s\n", iface );
    sniff.pDescr = pcap_open_live(iface, 65535, 0, 0, (char *)&pcaperr);
  }
  else if ( infile != NULL ) {
    fprintf( stderr, "[x] Offline from filename=%s\n", infile );
    sniff.pDescr = pcap_open_offline(infile, (char *)&pcaperr);
  }

  if ( !sniff.pDescr ) {
    fprintf(stderr, "Failed to open interface/file\n");
    exit(1);
  }

  if ( pcap_compile(sniff.pDescr, &bpf, pFilter, 0, 0) == -1 ) {
    fprintf(stderr, "pcap failed to compile filter: %s\n", pFilter );
    exit(1);
  }

  fprintf( stdout, "[x] Setting pcap filter ...\n");
  pcap_setfilter(sniff.pDescr, &bpf);
  pcap_freecode( &bpf );

  fprintf( stderr, "[x] Starting sniffer thread ...\n");
  pthread_create(&plugth, NULL, (void *)startsniffer, &sniff);

  pthread_join(plugth, NULL);

  if ( pcap_stats(sniff.pDescr, &ps ) != -1 ) {
    fprintf(stdout, "[x] Packets recieved: %d\n", ps.ps_recv);
    fprintf(stdout, "[x] Packets dropped: %d\n", ps.ps_drop);
  }

  pPlug = pPlugins;

  fprintf( stdout, "[x] Unloading plugins ...\n");
  while ( pPlug ) {
    nRecordCount += pPlug->plugin_packetcount();
    pPlug->plugin_stop();
    dlclose( pPlug->pHandle );
    fprintf( stdout, "[x] Succesfully unloaded %s\n", pPlug->name );
    pPlug = pPlug->next;
  }

  fprintf( stdout, "[x] Sucessfully unloaded plugins\n");
  fprintf(stdout, "[x] Disconnecting from provider ...\n");
  store_disconnect();
  fprintf(stdout, "[x] Disconnected from provider\n");
  fprintf(stdout, "[x] Succesfully stored %d records\n", nRecordCount );

  return 0;

}
