/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *
 * Plugin for detecting compaq Insight Manager
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "net.h"
#include "cdp.h"
#include "util.h"
#include "plugin.h"

struct host *pHost = NULL;
struct property *pProp = NULL;

int dupcheck(struct cdphost *list, struct cdphost *host) {

  struct cdphost *loop = list;

  while ( loop ) {
    
    if ( strcmp( loop->ether, host->ether ) == 0 )
      return 1;
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct cdphost *hosts ) {

  struct cdphost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct cdphost *pushost(struct cdphost *hosts, struct cdphost *newhost) {
  
  if ( dupcheck( hosts, newhost ) == 0x00 ) {
    newhost->next = hosts;
    hosts = newhost;
  }

  return hosts;

}

struct cdphost *createhost(char *ether, char *ip) {
  return NULL;
}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerbose ) {

  struct ether_header *ether;
  u_int8_t ciscover;
  u_int8_t *pMAC;
  u_int16_t cdpmsg_type;
  u_int16_t cdpmsg_len;
  u_int16_t *llcproto;
  int nOffset = 0, nDataLen = 0;
  char cdpdata[512];
  u_char *pData;

  struct cdphost *newhost;
  struct in_addr *tmpIp;
  
  ether = ( struct ether_header *)(pPacket);

  if ( nLen < sizeof( struct ether_header ) )
    return -1;

  pMAC = ether->ether_shost;

  /* ether header + llc cruft up to protocol ID */
  nOffset = sizeof( struct ether_header ) + 7;

  if ( nLen < nOffset )
    return -1;

  /* -1 because the proto is 16bit and offset points to second byte*/
  pData = (u_char *)(pPacket + nOffset - 1);
  llcproto = (u_int16_t *)pData;
  
  if ( *llcproto != htons(0x2000) )
    return -1;

  /* ver */
  nOffset += 1;

  if ( nLen < nOffset )
    return -1;

  pData = (u_char *)(pPacket + nOffset);

  ciscover = pData[0];

  /* ttl, chksum */
  nOffset += 4;

  /* 4 bytes type (16bit) length (16bit) */
  if ( nLen < nOffset + 4 )
    return -1;

  newhost = ( struct cdphost *) malloc ( sizeof ( struct cdphost ) );
  memcpy( newhost->ether, pMAC, sizeof( newhost->ether ) );

  while ( nOffset < nLen ) {

    pData = (u_char *)(pPacket + nOffset );
    cdpmsg_type = *(u_int16_t *)pData;
    cdpmsg_type = htons( cdpmsg_type );
    pData = (u_char *)(pPacket + nOffset + 2);
    cdpmsg_len = *(u_int16_t *)pData;
    cdpmsg_len = htons( cdpmsg_len );
    
    if ( nLen < ( nOffset + (cdpmsg_len) ) )
      return -1;

    /* type & length */
    nOffset += 4;
    cdpmsg_len -= 4;
    pData = (u_char *) ( pPacket + nOffset );

    if ( cdpmsg_len > sizeof( cdpdata ) )
	 return -1;

    memset( cdpdata, 0, sizeof( cdpdata ) );
    memcpy( cdpdata, pData, cdpmsg_len );

    /* device id */
    if ( cdpmsg_type == 0x0001 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->id ) - 1 < nDataLen )
	nDataLen = sizeof( newhost->id ) - 1;

      memset( newhost->id, 0, sizeof( newhost->id ));
      memcpy( newhost->id, cdpdata, nDataLen );

    }

    /* device id */
    if ( cdpmsg_type == 0x0002 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->address ) - 1 < nDataLen )
	nDataLen = sizeof( newhost->address ) - 1;

      memset( newhost->address, 0, sizeof( newhost->address ));
      memcpy( newhost->address, cdpdata, nDataLen );
      memset( newhost->ip, 0, sizeof( newhost->ip ) );

      /* do some lazy checks */
      /* is this an ip address ? */
      if ( nDataLen >= 13 && newhost->address[6] == 0xcc ) {
	/* is it 4 bytes long */
	if ( newhost->address[7] == 0x00 && newhost->address[8] == 0x04 ) {
	  tmpIp = (struct in_addr *)(&newhost->address[9]);
	  strncpy( newhost->ip, inet_ntoa( *tmpIp ), 
		   sizeof( newhost->ip ) - 1 );
	}
      }

    }

    /* port id */
    if ( cdpmsg_type == 0x0003 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->portid ) - 1 < nDataLen )
	nDataLen = sizeof( newhost->portid ) - 1;

      memset( newhost->portid, 0, sizeof( newhost->portid ));
      memcpy( newhost->portid, cdpdata, nDataLen );
    }

    /* sw version */
    if ( cdpmsg_type == 0x0005 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->ios ) - 1 < nDataLen )
	nDataLen = sizeof( newhost->ios ) - 1;

      memset( newhost->ios, 0, sizeof( newhost->ios ));
      memcpy( newhost->ios, cdpdata, nDataLen );
      //fprintf(stderr, "IOS: %s\n", newhost->ios );
    }

    /* platform */
    if ( cdpmsg_type == 0x0006 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->platform ) - 1 < nLen )
	nDataLen = sizeof( newhost->platform ) - 1;

      memset( newhost->platform, 0, sizeof( newhost->platform ));
      memcpy( newhost->platform, cdpdata, nDataLen );
      //fprintf(stderr, "Platform: %s\n", newhost->platform );
    }

    /* mgmt dom */
    if ( cdpmsg_type == 0x0009 ) {

      nDataLen = cdpmsg_len;

      if ( sizeof( newhost->mgmtdom ) - 1 < nLen )
	nDataLen = sizeof( newhost->mgmtdom ) - 1;

      memset( newhost->mgmtdom, 0, sizeof( newhost->mgmtdom ));
      memcpy( newhost->mgmtdom, cdpdata, nDataLen );
      //fprintf(stderr, "Mgmt. Dom.: %s\n", newhost->mgmtdom );
    }


    nOffset += cdpmsg_len;

  }

  cdphosts = pushost( cdphosts, newhost );
  plugin_storehost( newhost );

  return 0;
}

static int gethostcount() {

  struct cdphost* loop = cdphosts;
  int nHosts = 0;

  while ( loop ) {
    nHosts ++;
    loop = loop->next;
  }

  return nHosts;

}

char *plugin_getfilter() {
   return "ether[20]=0x20 && ether[21]=0x00";
}

void plugin_stats(FILE *pFD) {
  fprintf( pFD, "\n");
  fprintf( pFD, "---- CDP Plugin Statistics --------------------------\n");
  fprintf( pFD, "CDP Hosts: %d\n", gethostcount());
  fprintf( pFD, "-----------------------------------------------------\n");
}

void plugin_stop() {
  freehosts(cdphosts);
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_storehost(struct cdphost *h) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, h->ether);
  
  pHost = plug_storehost( pHost, mac, h->ip, h->id, "cdp");

  if ( !pHost ) {
    fprintf(stderr, "ERROR: An error occured while storing host\n");
    return;
  }
  else {
    nPacketCount ++;
    plug_storeproperty( pHost, "portid", h->portid );
    plug_storeproperty( pHost, "platform", h->platform );
    plug_storeproperty( pHost, "ios", h->ios );
  }

}

struct host *plugin_gethost() {
  return pHost;
}

struct property *plugin_getproperty() {
  return pProp;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

void plugin_freeproperty() {
  plug_freeproperty( pProp );
  pProp = NULL;
}

char *plugin_getdesc() {
  return "CDP - Cisco Discovery Protocol plugin";
}
