/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *
 * Plugin for detecting compaq Insight Manager
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "cim.h"
#include "util.h"
#include "plugin.h"
#include "net.h"

struct host *pHost = NULL;
struct property *pProp = NULL;

void plugin_dbstorehost(struct cimhost *);

void printhost(FILE *pFD, char *prefix, struct cimhost *host) {
  unsigned char mac[20];

  mactostr((unsigned char *)&mac, host->ether);
  fprintf(pFD, "%s,%s,%s\n", prefix, mac, host->ip );

}

int dupcheck(struct cimhost *list, struct cimhost *host) {

  struct cimhost *loop = list;

  while ( loop ) {
    
    if ( strcmp( loop->ip, host->ip ) == 0 )
      return 1;
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct cimhost *hosts ) {

  struct cimhost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct cimhost *pushost(struct cimhost *hosts, struct cimhost *newhost) {

  if ( dupcheck( hosts, newhost ) == 0x00 ) {
    newhost->next = hosts;
    hosts = newhost;
  }

  return hosts;

}

struct cimhost *createhost(char *ether, char *ip) {

  struct cimhost *newhost;

  newhost = (struct cimhost *) malloc ( sizeof( struct cimhost ) );
  memcpy( newhost->ether, ether, sizeof( newhost->ether ) );
  memcpy( newhost->ip, ip, sizeof( newhost->ip ) );

  plugin_dbstorehost( newhost );

  return newhost;

}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerbose ) {

  int nOffset;
  struct cimhost *newhost;
  struct in_addr *inetaddr;
  struct ip_header *pIpheader;
  struct udphdr *pUdpheader;
  struct ether_header *pEthheader;
  u_int8_t *pSrcEth;

  nOffset = sizeof( struct ether_header ) + sizeof( struct ip_header );

  if ( nLen < nOffset ) {
    return -1;
  }

  pEthheader = (struct ether_header *)pPacket;
  pSrcEth = pEthheader->ether_shost;

  pIpheader = (struct ip_header *)(pPacket + nOffset - sizeof( struct ip_header ) );
  inetaddr = (struct in_addr *)&pIpheader->ip_src;

  if ( pIpheader->ip_p != IPPROTO_UDP ) {
    return -1;
  }

  pUdpheader = (struct udphdr *)(pPacket + sizeof(struct ether_header) +
				 IP_HL(pIpheader) * 4 );

  nOffset += ntohs(pUdpheader->uh_ulen) - sizeof( struct udphdr );

  if ( nLen < nOffset ) {
    return -1;
  }

  if ( htons(pUdpheader->uh_sport) == 2301 ) {
    newhost = createhost( pSrcEth, inet_ntoa( *inetaddr  ) );

    if ( nVerbose )
      printhost( stderr, "cim", newhost );

    cimhosts = pushost( cimhosts, newhost );

    return 0;
  }

  return -1;

}

void plugin_report(FILE *pFD) {

  struct cimhost *loop = cimhosts;

  while ( loop ) {
    printhost(pFD, "cim", loop);
    loop = loop->next;
  }

}

static int gethostcount() {
  
  struct cimhost* loop = cimhosts;
  int nHosts = 0;

  while ( loop ) {
    nHosts ++;
    loop = loop->next;
  }

  return nHosts;

}

void plugin_stats(FILE *pFD) {
  fprintf( pFD, "\n");
  fprintf( pFD, "---- CIM Plugin Statistics --------------------------\n");
  fprintf( pFD, "Compaq Insight Manager Hosts: %d\n", gethostcount());
  fprintf( pFD, "-----------------------------------------------------\n");
}

void plugin_stop() {
  freehosts(cimhosts);
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_dbstorehost(struct cimhost *h) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, h->ether);
  pHost = plug_storehost( pHost, mac, h->ip, "", "cim");  

  if ( !pHost )
    return;

  nPacketCount ++;

}

struct host *plugin_gethost() {
  return pHost;
}

struct property *plugin_getproperty() {
  return pProp;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

void plugin_freeproperty() {
  plug_freeproperty( pProp );
  pProp = NULL;
}

char *plugin_getdesc() {
  return "CIM - Compaq Insight Manager plugin";
}
