/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _SMB_H_
#define _SMB_H_

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>

struct ip_header {
        u_int8_t        ip_vhl;         /* header length, version */
#define IP_V(ip)        (((ip)->ip_vhl & 0xf0) >> 4)
#define IP_HL(ip)       ((ip)->ip_vhl & 0x0f)
        u_int8_t        ip_tos;         /* type of service */
        u_int16_t       ip_len;         /* total length */
        u_int16_t       ip_id;          /* identification */
        u_int16_t       ip_off;         /* fragment offset field */
#define IP_DF 0x4000                    /* dont fragment flag */
#define IP_MF 0x2000                    /* more fragments flag */
#define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
        u_int8_t        ip_ttl;         /* time to live */
        u_int8_t        ip_p;           /* protocol */
        u_int16_t       ip_sum;         /* checksum */
        struct  in_addr ip_src,ip_dst;  /* source and dest address */
};

struct nb_datagram_svc {
  u_char msgtype;
  u_char frag;
  u_int16_t id;
  struct in_addr sip;
  u_int16_t sport;
  u_int16_t len;
  u_int16_t offset;
  char srcname[34];
  char dstname[34];
}__attribute__((packed));

struct netbios {
  u_int16_t length;
  u_int16_t delimiter;
  u_int8_t cmd;
  u_int8_t unknown[7];
  u_int8_t recvname[16];
  u_int8_t sendname[16];
} __attribute__((packed));

struct smb_header {
  u_char srvcomp[4];
  u_char cmd;
  u_char errclass;
  u_char reserved;
  u_int16_t errcode;
  u_char flags1;
  u_int16_t flags2;
  u_char reserved2[12];
  u_int16_t treeid;
  u_int16_t processid;
  u_int16_t userid;
  u_int16_t multiplexid;
} __attribute__((packed));

struct trans_req {
  u_char wordcount;
  u_int16_t totalparam;
  u_int16_t totaldata;
  u_int16_t maxparam;
  u_int16_t maxdata;
  u_char maxsetup;
  u_char reserved;
  u_int16_t flags;
  long timeout;
  u_int16_t reserved2;
  u_int16_t paramcount;
  u_int16_t paramoffset;
  u_int16_t datacount;
  u_int16_t dataoffset;
  u_char setupcount;
  u_char reserved3;
} __attribute__((packed));

struct smb_mailslot {
  u_int16_t opcode;
  u_int16_t priority;
  u_int16_t class;
  u_int16_t size;
} __attribute__((packed));

/* host announcement */
struct ms_browseproto_ha {
  u_char cmd;
  u_char updatecount;
  long updateperiod;
  char hostname[16];
  u_char osmajor;
  u_char osminor;
  long servertype;
  u_char protomajor;
  u_char protominor;
  u_int16_t signature;
  /*  char comment[64]; */
} __attribute__((packed));

/* domain announcement */
struct ms_browseproto_da {
  u_char cmd;
  u_char updatecount;
  long updateperiod;
  char domain[16];
  u_char osmajor;
  u_char osminor;
  long servertype;
  long mysterious;
  /*char masterbrowser[16];*/
} __attribute__((packed));
struct nb_ns_hdr {
  u_int16_t trans;
  u_int16_t flags;
  u_int16_t questions;
  u_int16_t answers;
  u_int16_t authority;
  u_int16_t rrs;
} __attribute__((packed));

struct nb_ns_query {
  char name[34];
  u_int16_t type;
  u_int16_t class;
} __attribute__((packed));

struct nb_ns_record {
  u_int16_t name;
  u_int16_t type;
  u_int16_t class;
  long ttl;
  u_int16_t length;
  u_int16_t flags;
  struct in_addr address;
} __attribute__((packed));

struct smbhost {
  u_int16_t proto;
  u_int8_t ether[6];
  struct in_addr ip;
  char nbname[32];
  u_char osmajor;
  u_char osminor;
  long srvtype;
  char comment[32];
  int type;
  u_int8_t md5[16];
  void *next;
};

struct smbdomain {
  u_int8_t ether[6];
  struct in_addr ip;
  char domain[16];
  char masterbrowser[32];
  void *next;
};

static struct smbhost *smbhosts = NULL;
static struct smbdomain *smbdoms = NULL;
static char *pMask = "(udp && (port 138 || port 137)) or (ether proto 0xf0)"; 

static int nVerbose;
static u_int16_t nPacketCount = 0;

void printhost(FILE *, char *, struct smbhost *);
void printdomain(FILE *, char *, struct smbdomain *);
struct smbhost *pushhost(struct smbhost *, struct smbhost *);

const int NB_HOST = 0x01;
const int SMB_HOST = 0x02;

#define SV_TYPE_WORKSTATION         0x00000001
#define SV_TYPE_SERVER              0x00000002
#define SV_TYPE_SQLSERVER           0x00000004
#define SV_TYPE_DOMAIN_CTRL         0x00000008
#define SV_TYPE_DOMAIN_BAKCTRL      0x00000010
#define SV_TYPE_TIME_SOURCE         0x00000020
#define SV_TYPE_AFP                 0x00000040
#define SV_TYPE_NOVELL              0x00000080
#define SV_TYPE_DOMAIN_MEMBER       0x00000100
#define SV_TYPE_PRINTQ_SERVER       0x00000200
#define SV_TYPE_DIALIN_SERVER       0x00000400
#define SV_TYPE_XENIX_SERVER        0x00000800
#define SV_TYPE_SERVER_UNIX         SV_TYPE_XENIX_SERVER
#define SV_TYPE_NT                  0x00001000
#define SV_TYPE_WFW                 0x00002000
#define SV_TYPE_SERVER_MFPN         0x00004000
#define SV_TYPE_SERVER_NT           0x00008000
#define SV_TYPE_POTENTIAL_BROWSER   0x00010000
#define SV_TYPE_BACKUP_BROWSER      0x00020000
#define SV_TYPE_MASTER_BROWSER      0x00040000
#define SV_TYPE_DOMAIN_MASTER       0x00080000
#define SV_TYPE_SERVER_OSF          0x00100000
#define SV_TYPE_SERVER_VMS          0x00200000
#define SV_TYPE_WINDOWS             0x00400000  /* Windows95 and above */
#define SV_TYPE_DFS                 0x00800000  /* Root of a DFS tree */
#define SV_TYPE_CLUSTER_NT          0x01000000  /* NT Cluster */
#define SV_TYPE_TERMINALSERVER      0x02000000
#define SV_TYPE_DCE                 0x10000000  /* IBM DSS (Directory and Securi
ty Services) or equivalent */
#define SV_TYPE_ALTERNATE_XPORT     0x20000000  /* return list for alternate tra
nsport */
#define SV_TYPE_LOCAL_LIST_ONLY     0x40000000  /* Return local list only */
#define SV_TYPE_DOMAIN_ENUM         0x80000000
#define SV_TYPE_ALL                 0xFFFFFFFF  /* handy for NetServerEnum2 */

#endif
