/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <string.h>
#include <pcap.h>
#ifdef NET_ETHERNET_H
#include <net/ethernet.h>
#endif
#include <netinet/in.h>
#ifdef NETINET_IP_H
#include <netinet/ip.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include "ipx.h"
#include "util.h"
#include "plugin.h"

static u_int16_t SAP_SOCKET = 0x0452;

void plugin_dbstorehost( struct nwhost *);

struct host *pHost = NULL;

char *plugin_getfilter() {
  /* return "ipx"; */ /* ipx doesnt work on older libpcap */
  return "ether[30]==0x04 && ether[31]==0x52 && ether[44]==0x00 && ether[45]==0x02";
}

void gethosttype( char desc[1024], const u_int16_t srvtype ) {

  if ( srvtype == SRV_TYPE_FILESERVER )
    strncpy(desc, "File Server", 1024 -1 );
  else if ( srvtype == SRV_TYPE_UNIVELUNIXWARE )
    strncpy(desc, "Univel Unixware", 1024 -1 );
  else if ( srvtype == SRV_TYPE_HPJETDIRECT )
    strncpy(desc, "HP Jetdirect", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_LANTERN )
    strncpy(desc, "LANtern", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NWMANAGEMENTAGENT )
    strncpy(desc, "NW Management Agent", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_MSGWSERVICESNW )
    strncpy(desc, "MS Gateway Services for NW", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NDS4 )
    strncpy(desc, "DS (NetWare 4.x)", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_BTRIEVENLM5 )
    strncpy(desc, "Btrieve VAP/NLM 5.0", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_SMSTESTDEVELOP )
    strncpy(desc, "SMS Testing and Development", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_TESNWFORVMS )
    strncpy(desc, "TES-Netware for VMS", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_LANPROTECTBINDERY )
    strncpy(desc, "LAN Protect Bindery", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NW386ORRSPXRCONSOLE )
    strncpy(desc, "NetWare 386 or RSPX Remote Console", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_ARCSERVE501 )
    strncpy(desc, "ARCServe 5.01", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_TIMESYNCSERVER )
    strncpy(desc, "Time Sync Server (NetWare 4.x)", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_ADVPRINTSVC )
    strncpy(desc, "Advertising Print Server", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NOVELLTIRPC )
    strncpy(desc, "Novell TI-RPC", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_XEROX )
    strncpy(desc, "XEROX", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_KYOCERA )
    strncpy(desc, "Kyocera", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_MSIIS )
    strncpy(desc, "MS IIS", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_CANONGP55 )
    strncpy(desc, "Canon GP 55 Printer", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_LEGATO )
    strncpy(desc, "Legato", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_LOTUSNOTES )
    strncpy(desc, "Lotus Notes", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NW386 )
    strncpy(desc, "Netware 386", 1024 - 1 );
  else if ( srvtype == 0x0077 )
    strncpy(desc, "Unknown", 1024 - 1 );
  else if ( srvtype == SRV_TYPE_NMSIPDISCORLANTERNTRAP )
    strncpy(desc, "NMS IP Disc. or LANtern Trap/Alarm", 1024 - 1 );
  else
    snprintf(desc, 1024 -1, "0x%X", srvtype);

}


void printhost(FILE *pFD, char *prefix, struct nwhost *host) {

  unsigned char mac[20];
  char service[1024];

  mactostr((unsigned char *)&mac, host->ether);

  /* fix names of time serves and ds servers */
  if ( ( host->srvtype == SRV_TYPE_NDS4 ) || 
       ( host->srvtype == SRV_TYPE_TIMESYNCSERVER ) )
    host->server[32] = 0;

  fprintf(pFD, "%s,%s,%s", prefix, mac, host->server);
  memset( service, 0, sizeof( service ) );
  gethosttype((char *)&service, host->srvtype);
  fprintf(pFD, "%s", service );  

  fprintf(pFD, "\n");

}

int dupcheck(const struct nwhost *list, 
	     const u_char md5dig[16] ) {

  struct nwhost *loop = (struct nwhost *)list;

  while ( loop ) {
    
    if ( md5comp( loop->md5, md5dig ) == 0 ) {
      return 1;
    }
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct nwhost *hosts ) {

  struct nwhost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct nwhost *pushost(struct nwhost *hosts, struct nwhost *newhost) {


  plugin_dbstorehost( newhost );
  newhost->next = hosts;
  hosts = newhost;

  return hosts;

}


struct nwhost *createhost(const char ether[6], 
			  const char servername[48], 
			  const u_int16_t srvtype,  
			  const u_int8_t network[4], 
			  const u_int8_t node[6]) {

  struct nwhost *newhost;
  u_char *md5dig;
  char data[128];
  int nLen;

  nLen = snprintf( data, sizeof( data ), "%s%d", servername, srvtype );
  md5dig = createMD5( data, nLen );

  if ( dupcheck( nwhosts, md5dig ) > 0 ) {
    free( md5dig );
    return NULL;
  }

  newhost = (struct nwhost *) malloc ( sizeof( struct nwhost ) );

  if ( !newhost ) {
    free( md5dig );
    return NULL;
  }

  memcpy( newhost->md5, md5dig, sizeof( newhost->md5 ) );
  free( md5dig );
  
  memcpy( (char *)&newhost->server, servername, sizeof( newhost->server ) - 1);
  memcpy( newhost->ether, ether, sizeof( newhost->ether ) );
  memcpy( (u_char *)&newhost->network, network, sizeof( newhost->network ) );
  memcpy( (u_char *)&newhost->node, node, sizeof( newhost->node ) );
  
  newhost->srvtype = srvtype;

  return newhost;

}

int plugin_manglepacket( const u_char *pPacket, const int nLen, 
			 const int nVerbose ) {

  struct ipx_header *ipxhdr;
  struct ipx_sap *ipxsap;
  struct nwhost *newhost;
  struct ether_header *pEthheader;
  u_int8_t *pSrcEth;
  int nOffset;

  /* OpenBSD hardcore kludge */
  nOffset = sizeof( struct ether_header ) + sizeof( struct ipx_header );

  if ( nLen < nOffset ) {
    fprintf(stderr, "Launch the missiles and call the president\n");
    return -1;
  }

  pEthheader = (struct ether_header *)pPacket;
  pSrcEth = pEthheader->ether_shost;

  ipxhdr = (struct ipx_header *)( pPacket + nOffset - sizeof( struct ipx_header ));

  if ( htons(ipxhdr->dstsocket) == SAP_SOCKET ) {

    nOffset += 2;

    if ( nLen < nOffset ) {
      return -1;
    }

    /* check for general response */
    if ( *( pPacket + nOffset - 1) != 0x02 ) {
      return -1;
    }

    while ( nLen >= nOffset ) {

      nOffset += sizeof( struct ipx_sap );

      if ( nLen < nOffset ) {
	break;
      }
    
      ipxsap = (struct ipx_sap *)(pPacket + nOffset - sizeof( struct ipx_sap ) );

      newhost = createhost(pSrcEth, ipxsap->servername, htons(ipxsap->srvtype),
			   ipxsap->network, ipxsap->node);

      if ( nVerbose )
	printhost( stderr, "novell", newhost );

      if ( newhost != NULL )
	nwhosts = pushost( nwhosts, newhost );
      
    }

    return 0;

  }

  return -1;

}

void plugin_report(FILE *pFD) {

  struct nwhost *loop = nwhosts;

  while ( loop ) {
    printhost(pFD, "novell", loop);
    loop = loop->next;
  }

}

static int gethostsoftype(const u_int16_t nType) {
  
  struct nwhost* loop = nwhosts;
  int nHosts = 0;

  while ( loop ) {

    if ( loop->srvtype == nType )
      nHosts ++;

    loop = loop->next;
  }

  return nHosts;

}

void plugin_stats(FILE *pFD) {
  fprintf( pFD, "\n");
  fprintf( pFD, "---- IPX Plugin Statistics --------------------------\n");
  fprintf( pFD, "NW File Servers: %d\n", gethostsoftype(SRV_TYPE_FILESERVER));
  fprintf( pFD, "-----------------------------------------------------\n");
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_stop() {
  freehosts(nwhosts);
}

void ipxaddrtostr( u_char *ipx, const u_char net[4], const u_char node[6] ) {

  int i = 0;
  u_char ipxnet[9];
  u_char ipxnode[13];

  memset( ipx, 0, 20 );
  memset( ipxnet, 0, sizeof( ipxnet ) );
  memset( ipxnode, 0, sizeof( ipxnode ) );

  for ( i=0; i<4; i ++ ) {
    snprintf((u_char *)&ipxnet + (i*2), sizeof( ipxnet ) - i*2, "%2X", net[i]);
    if ( ipxnet[i*2] == ' ' )
      ipxnet[i*2]='0';
  }


  for ( i=0; i<6; i ++ ) {
    snprintf((u_char *)&ipxnode + (i*2), sizeof( ipxnode ) - i*2, 
	     "%2X", node[i] );
    if ( ipxnode[i*2] == ' ' )
      ipxnode[i*2]='0';
  }

  strncpy( ipx, ipxnet, 8 );
  strncat( ipx, ":", 1 );
  strncat( ipx, ipxnode, 12 );

}

void plugin_dbstorehost( struct nwhost *h ) {

  unsigned char mac[20];
  unsigned char ipxaddr[22];
  char service[1024];

  mactostr((unsigned char *)&mac, h->node);
  ipxaddrtostr( (char *)&ipxaddr, h->network, h->node );
  memset( service, 0, sizeof( service ) );

  if ( h->srvtype > 0 )
    gethosttype( (char *)&service, h->srvtype );


  /* fix names of time serves and ds servers */
  if ( ( h->srvtype == SRV_TYPE_NDS4 ) || 
       ( h->srvtype == SRV_TYPE_TIMESYNCSERVER ) )
    h->server[32] = 0;
  
  pHost = plug_storehost( pHost, mac, ipxaddr, h->server, "ipxhost");

  if ( !pHost )
    return;

  nPacketCount ++;

  if ( h->srvtype > 0 )
    plug_storeproperty( pHost, "srvtype", service);


}

struct host *plugin_gethost() {
  return pHost;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

char *plugin_getdesc() {
  return "IPX - IPX SAP Plugin";
}
