/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 *
 * Plugin for detecting broadcasting tftp requests
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "net.h"
#include "tftp.h"
#include "util.h"
#include "plugin.h"

struct host *pHost = NULL;

void plugin_dbstorehost(struct tftphost *);

void printhost(FILE *pFD, char *prefix, struct tftphost *host) {
  unsigned char mac[20];

  mactostr((unsigned char *)&mac, host->ether);
  fprintf(pFD, "%s,%s,%s\n", prefix, mac, host->ip );

}

int dupcheck(struct tftphost *list, struct tftphost *host) {

  struct tftphost *loop = list;

  while ( loop ) {
    
    if ( strcmp( loop->ip, host->ip ) == 0 && 
	 strcmp( loop->filename, host->filename) == 0 )
      return 1;
    
    loop = loop->next;

  }

  return 0;

}

/*
 * Frees memory of all hosts in linked list
 *
 */
void freehosts( struct tftphost *hosts ) {

  struct tftphost *curr, *rmv;

  curr = hosts;

  while ( curr ) {
    rmv = curr;
    curr = curr->next;
    free(rmv);
  }

}

struct tftphost *pushost(struct tftphost *hosts, struct tftphost *newhost) {

  if ( dupcheck( hosts, newhost ) == 0x00 ) {
    plugin_dbstorehost( newhost );
    newhost->next = hosts;
    hosts = newhost;
  }

  return hosts;

}

struct tftphost *createhost(char *ether, char *ip, char *filename) {

  struct tftphost *newhost;

  newhost = (struct tftphost *) malloc ( sizeof( struct tftphost ) );
  memcpy( newhost->ether, ether, sizeof( newhost->ether ) );
  memcpy( newhost->ip, ip, sizeof( newhost->ip ) );
  strncpy( newhost->filename, filename, sizeof( newhost->filename ) - 1 );

  return newhost;

}

int plugin_manglepacket( const u_char *pPacket, int nLen, int nVerbose ) {

  int nOffset;
  struct tftphost *newhost;
  struct in_addr *inetaddr;
  struct ip_header *pIpheader;
  struct udphdr *pUdpheader;
  struct ether_header *pEthheader;
  u_int8_t *pSrcEth;
  u_int8_t *pFilename;
  u_int8_t *pData;
  int i;

  nOffset = sizeof( struct ether_header ) + sizeof( struct ip_header );

  if ( nLen < nOffset ) {
    return -1;
  }

  pEthheader = (struct ether_header *)pPacket;
  pSrcEth = pEthheader->ether_shost;

  pIpheader = (struct ip_header *)(pPacket + nOffset - sizeof( struct ip_header ) );
  inetaddr = (struct in_addr *)&pIpheader->ip_src;

  if ( pIpheader->ip_p != IPPROTO_UDP ) {
    return -1;
  }

  pUdpheader = (struct udphdr *)(pPacket + sizeof(struct ether_header) +
				 IP_HL(pIpheader) * 4 );

  nOffset += sizeof( struct udphdr );

  if ( nLen < nOffset ) {
    return -1;
  }

  if ( htons(pUdpheader->uh_dport) == 69 ) {

    if ( nLen < nOffset + 2 )
      return -1;

    pData = ( u_int8_t *)( pPacket + nOffset );

    /* tftp read request */
    if ( pData[1] != 0x01 )
      return -1;

    nOffset += 2;
    pFilename = ( u_int8_t *) ( pPacket + nOffset );

    i = nOffset;

    while ( pFilename[i] != 0 && i<nLen )
      i++;

    if ( i==nLen )
      return -1;

    newhost = createhost( pSrcEth, inet_ntoa( *inetaddr  ), pFilename );
    tftphosts = pushost( tftphosts, newhost );

    return 0;
  }

  return -1;

}

void plugin_report(FILE *pFD) {

  struct tftphost *loop = tftphosts;

  while ( loop ) {
    printhost(pFD, "cim", loop);
    loop = loop->next;
  }

}

static int gethostcount() {
  
  struct tftphost* loop = tftphosts;
  int nHosts = 0;

  while ( loop ) {
    nHosts ++;
    loop = loop->next;
  }

  return nHosts;

}

void plugin_stats(FILE *pFD) {

}

void plugin_stop() {
  freehosts(tftphosts);
}

int plugin_packetcount() {
  return nPacketCount;
}

void plugin_dbstorehost(struct tftphost *h) {

  unsigned char mac[20];

  mactostr((unsigned char *)&mac, h->ether);
  pHost = plug_storehost( pHost, mac, h->ip, "", "tftp");

  if ( !pHost )
    return;

  plug_storeproperty( pHost, "filename", h->filename );
  nPacketCount ++;

}

struct host *plugin_gethost() {
  return pHost;
}

void plugin_freehost() {
  plug_freehost( pHost );
  pHost = NULL;
}

char *plugin_getfilter() {
  return "udp && port 69";
}

char *plugin_getdesc() {
  return "TFTP - Trivial File Transfer Protocol plugin";
}
