/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "util.h"

static int nSessionId = 0;

char *db_parameters() {
  return "filename=<filename>;session=<session>";
}

int db_storehost( FILE *pConn, const char *mac, const char *addr, 
		  const char *name, const char *type, const char *chksum ) {
  fprintf( pConn, "insert into hosts( session_id, mac, addr, name, type, chksum) values(%d, '%s','%s','%s','%s','%s');\n", 
	   nSessionId, mac, addr, name, type, chksum );

  return 0;
}

int db_storeproperty(FILE *pConn, const char *name,
		     const char *val, const char *chksum) {
  fprintf( pConn, "insert into property( host_chksum, prop, val ) values('%s','%s','%s');\n", chksum, name, val );

  return 0;
}

void db_createsession( FILE *pFD, char *session ) {
  nSessionId ++;

  fprintf( pFD, 
	   "insert into session ( id, name, created ) values(%d,'%s','%s');\n",
	   nSessionId, session, "2002-01-01" );
}

FILE *db_connect( const char *pConnURL ) {

  FILE *pFD;
  char *pFilename = getparamvalue( pConnURL, "filename" );
  char *pSession = getparamvalue( pConnURL, "session" );
 
  if ( !pFilename || !pSession ) {
    fprintf( stderr, "Connect failed make sure all parameters are supplied\n");
    fprintf( stderr, "%s\n", db_parameters() );
    freeparamvalue( pFilename );
    return NULL;
  }

  pFD = fopen( pFilename, "a" );

  if ( pFD )
    db_createsession( pFD, (char *)pSession );

  freeparamvalue( pFilename );
  freeparamvalue( pSession );

  return pFD;

}

void db_disconnect(FILE *pConn) {
  if ( pConn )
    fclose( pConn );
}

char *db_getname() {
  return "GENSQL";
}

char *db_getdescription() {
  return "Generic SQL Plugin that creates a SQL script";
}
