/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <tds.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "util.h"

static int nSession;
TDSCONTEXT *m_pContext;
TDSLOGIN *m_pLogin;

char *db_parameters( ) {

  char *pParams = 
    "user=<user>;pass=<pw>;db=<dbname>;host=<host>;session=<session>";

  return pParams;

}

int tsql_handle_message( TDSCONTEXT *pContext, 
			 TDSSOCKET *pTds,
			 TDSMSGINFO *pMsg ) {
  
  if (pMsg->msg_number == 0) {
    fprintf(stderr, "%s\n", pMsg->message);
    return 0;
  }

  if (pMsg->msg_number != 5701 && pMsg->msg_number != 20018) {
    fprintf(stderr, "Msg %d, Level %d, State %d, Server %s, Line %d\n%s\n",
	    pMsg->msg_number, pMsg->msg_level, pMsg->msg_state,
	    pMsg->server, pMsg->line_number, pMsg->message);
  }

  return -1;

}

int db_createsession( TDSSOCKET *pConn, const char *session ) {

  TDS_INT resulttype;
  char *SQLfrag = "insert into session(id, name, created) values(";
  char sSQL[1024];
  int nErr = 0, nErr2 = 0;
  int nLen, i=0;

  nLen = strlen( session );

  if ( nLen > sizeof( sSQL ) ) {
    fprintf( stderr, "ERROR: Failed to store host\n");
    return -1;
  }

  nSession = 1;

  do {

    if ( nSession > 1023 )
      break;

    snprintf( (char *)&sSQL, sizeof(sSQL)-1, "%s%d,'%s', getdate())", SQLfrag, nSession, session );

    nErr = tds_submit_query(pConn, sSQL );

    if ( nErr != TDS_SUCCEED ) {
      fprintf( stderr, "FAIL!\n");
      nSession ++;
    }

    while ((nErr2 = tds_process_result_tokens(pConn, &resulttype)) == TDS_SUCCEED)
      i++;

  } while ( nErr != TDS_SUCCEED );

  return nErr;

}

/*
 * Stores host
 *
 */
int db_storehost( TDSSOCKET *pConn, char *mac, char *addr, char *name, 
		   char *type, char *chksum ) {

  TDS_INT resulttype;

  char *SQLfrag = "insert into hosts(session_id, mac, addr, name, type, chksum) values(";
  char sSQL[1024];
  int nErr = 0;
  int nLen, i=0;

  nLen = strlen( mac ) + strlen( addr ) + strlen( name ) + strlen( type ) +
    strlen( chksum );

  if ( nLen > sizeof( sSQL ) ) {
    return -1;
  }

  bzero( sSQL, sizeof( sSQL ) );
  snprintf( (char *)&sSQL, sizeof( sSQL ) - 1, 
	    "%s%d,'%s','%s','%s','%s', '%s')", 
	    SQLfrag, nSession, mac, addr, name, type, chksum);

  nErr = tds_submit_query(pConn, sSQL );

  if ( nErr != TDS_SUCCEED )
    return nErr;
  
  while ((nErr = tds_process_result_tokens(pConn, &resulttype)) == TDS_SUCCEED)
    i++;

  return 0;

}

int db_storeproperty(TDSSOCKET *pConn, const char *name, 
		     const char *val, const char *chksum ) {

  TDS_INT resulttype;
  char *SQLfrag = "insert into property(host_chksum,prop,val) "
                  "values(";
  char sSQL[1024];
  int nErr = 0, i=0;

  if ( strlen(name) + strlen(val) + strlen( SQLfrag ) > sizeof( sSQL ) )
    return -1;

  bzero( sSQL, sizeof( sSQL ) );
  snprintf( (char *)&sSQL, sizeof( sSQL )-1,
	    "%s'%s','%s','%s')", SQLfrag, chksum, name, val);

  nErr = tds_submit_query(pConn, sSQL );

  if ( nErr != TDS_SUCCEED ) 
    return -1;

  while ((nErr = tds_process_result_tokens(pConn, &resulttype)) == TDS_SUCCEED)
    i++;

  return 0;

}

TDSSOCKET *db_connect( char *pConnUrl ) {

  TDSSOCKET *pTds = NULL;
  TDSCONNECTINFO *pConnInfo;
 
  char *user, *pw, *db, *session, *host;

  m_pLogin = tds_alloc_login();
  m_pContext = tds_alloc_context();

  if (m_pContext->locale && !m_pContext->locale->date_fmt) {
    /* set default in case there's no locale file */
    m_pContext->locale->date_fmt = strdup("%b %e %Y %I:%M%p");
  }

  m_pContext->msg_handler = tsql_handle_message;
  m_pContext->err_handler = tsql_handle_message;

  user    = getparamvalue( pConnUrl, "user" );
  pw      = getparamvalue( pConnUrl, "pw" );
  db      = getparamvalue( pConnUrl, "db" );
  session = getparamvalue( pConnUrl, "session" );
  host    = getparamvalue( pConnUrl, "host" );

  if ( !user || !db || !session || !host) {
    fprintf( stderr, "Connect failed make sure all parameters are supplied\n");
    fprintf( stderr, "%s\n", db_parameters() );
  }
  else {

    tds_set_user( m_pLogin, user );
    tds_set_app( m_pLogin, "TSQL" );
    tds_set_library( m_pLogin, "TDS-Library" );
    tds_set_server( m_pLogin, host );
    //    tds_set_charset( m_pLogin, NULL );
    tds_set_language(m_pLogin, "us_english");
    tds_set_packet(m_pLogin, 512);
    tds_set_passwd(m_pLogin, pw);
    
    pTds = tds_alloc_socket(m_pContext, 512);
    tds_set_parent(pTds, NULL);
    pConnInfo = tds_read_config_info( NULL, m_pLogin, m_pContext->locale );
    
    if (!pConnInfo || tds_connect(pTds, pConnInfo) == TDS_FAIL) {
      tds_free_connect(pConnInfo);
      pTds = NULL;
      goto quit;
    }
    else {
      tds_free_connect( pConnInfo );
      db_createsession( pTds, session );
    }

  }

 quit:
  freeparamvalue( user );
  freeparamvalue( pw );
  freeparamvalue( db );
  freeparamvalue( session );
  freeparamvalue( host );

  return pTds;

}

void db_disconnect(TDSSOCKET *pTds) {

  tds_free_socket(pTds);
  tds_free_login(m_pLogin);
  tds_free_context(m_pContext);

}

char *db_getname() {
  return "MSSQL";
}

char *db_getdescription() {
  return "Storage plugin for Microsoft Sql Server";
}
