/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <mysql/mysql.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "util.h"

static int nSession;

char *db_parameters( ) {

  char *pParams = 
    "user=<user>;pass=<pw>;db=<dbname>;host=<host>;session=<session>";

  return pParams;

}


int db_createsession( MYSQL *pConn, const char *session ) {

  char *SQLfrag = "insert into session(id, name, created) values(";
  char sSQL[1024];
  int nErr = 0;
  int nLen;

  nLen = strlen( session );

  if ( nLen > sizeof( sSQL ) ) {
    fprintf( stderr, "ERROR: Failed to store host\n");
    return -1;
  }

  nSession = 1;

  do {

    if ( nSession > 1023 )
      break;

    snprintf( (char *)&sSQL, sizeof(sSQL)-1, "%s%d,'%s', now())", SQLfrag, nSession, session );

    nErr = 0;

    if ( mysql_real_query( pConn, sSQL, strlen( sSQL )) != 0 )
      nErr = mysql_errno( pConn );

    if ( nErr != 0 )
      nSession ++;

  } while ( nErr != 0 );

  return nErr;

}

/*
 * Stores host
 *
 */
int db_storehost( MYSQL *pConn, char *mac, char *addr, char *name, 
		   char *type, char *chksum ) {

  char *SQLfrag = "insert into hosts(session_id, mac, addr, name, type, chksum) values(";
  char sSQL[1024];
  int nErr = 0;
  int nLen;

  nLen = strlen( mac ) + strlen( addr ) + strlen( name ) + strlen( type ) +
    strlen( chksum );

  if ( nLen > sizeof( sSQL ) ) {
    fprintf( stderr, "ERROR: Failed to store host\n");
    return -1;
  }

  snprintf( (char *)&sSQL, sizeof( sSQL ) - 1, 
	    "%s%d,'%s','%s','%s','%s', '%s')", 
	    SQLfrag, nSession, mac, addr, name, type, chksum);

  if ( mysql_real_query( pConn, sSQL, strlen( sSQL )) != 0 )
    nErr = mysql_errno( pConn );

  /* 1062 => duplicate detected */
  if (nErr != 0 && nErr != 1062)
    fprintf(stderr, "error on query: %d (%s)\n", mysql_errno(pConn), sSQL);

  if ( nErr == 1062 )
    return 1111; /* dup */

  return nErr;
  
}

int db_storeproperty(MYSQL *pConn, const char *name, const char *val, 
		     const char *chksum) {

  char *SQLfrag = "insert into property(host_chksum,prop,val) "
                  "values(";
  char sSQL[1024];
  int nErr = 0;
  
  if ( strlen(name) + strlen(val) + strlen( SQLfrag ) > sizeof( sSQL ) )
    return -1;

  snprintf( (char *)&sSQL, sizeof( sSQL )-1,
	    "%s'%s','%s','%s')", SQLfrag, chksum, name, val);

  if ( mysql_real_query( pConn, sSQL, strlen( sSQL )) != 0 )
    nErr = mysql_errno( pConn );

  /* 1062 => duplicate detected */
  if (nErr != 0 && nErr != 1062)
    fprintf(stderr, "error on query: %d\n", mysql_errno( pConn ));

  return nErr;

}

MYSQL *db_connect( char *pConnUrl ) {

  char *user, *pw, *db, *session, *host;
  MYSQL *pConn = malloc( sizeof( MYSQL ) );

  if (mysql_init( pConn ) == NULL)
    goto fail;

  if (mysql_options( pConn, MYSQL_READ_DEFAULT_GROUP, "simple") != 0)
    goto fail;

  user    = getparamvalue( pConnUrl, "user" );
  pw      = getparamvalue( pConnUrl, "pw" );
  db      = getparamvalue( pConnUrl, "db" );
  session = getparamvalue( pConnUrl, "session" );
  host    = getparamvalue( pConnUrl, "host" );

  if ( !user || !db || !session ) {
    fprintf( stderr, "Connect failed make sure all parameters are supplied\n");
    fprintf( stderr, "%s\n", db_parameters() );
    free( pConn );
    pConn = NULL;
  }
  else {

    if (mysql_real_connect( pConn, host, user, pw, db, 0, NULL, 0) == NULL) {
      fprintf(stderr, "error on connect: %s\n", mysql_error( pConn ));
    }
    else
      db_createsession( pConn, session );

  }

  freeparamvalue( user );
  freeparamvalue( pw );
  freeparamvalue( db );
  freeparamvalue( session );
  freeparamvalue( host );

  return pConn;

 fail:
  free( pConn );
  return NULL;

}

MYSQL_RES *db_getresult( MYSQL *pConn, const char *pQuery ) {

  MYSQL_RES *pRes;

  if ( mysql_real_query(pConn, pQuery, strlen( pQuery )) != 0 ) {
    fprintf( stderr, "Query failed\n");
  }

  if ((pRes = mysql_store_result(pConn)) == NULL) {
    fprintf(stderr, "error on store: %s\n", mysql_error(pConn));
    return NULL;
  }

  return pRes;

}

MYSQL_ROW db_nextrow( MYSQL_RES *pRes ) {
  
  MYSQL_ROW row;

  if ( ( row = mysql_fetch_row(pRes) ) ) {
    return row;
  }
  else
    return NULL;

}

void db_freeresult( MYSQL_RES *pRes ) {
  if ( pRes )
    mysql_free_result(pRes);
}

void db_disconnect(MYSQL *pConn) {
  if ( pConn ) {
    mysql_close(pConn);
    free(pConn);
    nSession = 0;
  }
}

char *db_getname() {
  return "MYSQL";
}

char *db_getdescription() {
  return "Storage plugin for MYSQL RDBMS";
}
