/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <string.h>
#include <dirent.h>
#include "store.h"
#include "plugin.h"
#include "util.h"
#include "errors.h"

static void *pConn = NULL;
struct store_info *pInfo;

int store_loadplug( char *pPlug ) {

  char *error;

  pInfo = (struct store_info *) malloc ( sizeof( struct store_info ) );

  pInfo->handle = dlopen( pPlug, RTLD_LAZY);

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s", pPlug );
    return 0;
  }

  pInfo->db_storeproperty = dlsym(pInfo->handle, "db_storeproperty");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  pInfo->db_storehost = dlsym(pInfo->handle, "db_storehost");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  pInfo->db_connect = dlsym(pInfo->handle, "db_connect");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  pInfo->db_disconnect = dlsym(pInfo->handle, "db_disconnect");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  pInfo->db_getname = dlsym(pInfo->handle, "db_getname");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  pInfo->db_getdescription = dlsym(pInfo->handle, "db_getdescription");

  if ( (error = (char *)dlerror()) != NULL) {
    fprintf(stderr, "%s\n", error );
    return 0;
  }

  return 1;

}

void store_unloadplug( ) {
  if ( !pInfo ) {
    dlclose( pInfo->handle );
    free( pInfo );
  }
}

char *store_getplugdesc( char *pPlug ) {

  char *pPlugInfo;
  int nSize = 1024;
  int nLen;

  pPlugInfo = (char *) malloc( nSize );
  bzero( pPlugInfo, nSize );

  if ( store_loadplug( pPlug ) ) {
    strncat( pPlugInfo, pInfo->db_getname(), nSize - 1 );
    nLen = nSize - strlen( pPlugInfo ) - 1;

    if ( nLen > 4 )
      strcat( pPlugInfo, " - " );

    nLen -= 4;
    strncat( pPlugInfo, pInfo->db_getdescription(), nLen );

    store_unloadplug();
  }

  return pPlugInfo;

}

char *store_getplugbyname( char *pName ) {

  DIR *pDir;
  struct dirent *pEntry;
  int nSize = 50;
  char *pluginname;

  pDir = opendir("./db");

  pluginname = (char *) malloc ( nSize );

  while ( ( pEntry = readdir( pDir ) ) ) {

    if ( strcmp(pEntry->d_name, "..") != 0 &&
	 strcmp(pEntry->d_name, "." ) != 0 ) {

      bzero( pluginname, nSize );
      strcpy( pluginname, "db/" );
      strncat( pluginname, pEntry->d_name, 20 );

      if ( store_loadplug( pluginname ) ) {

	if ( strcmp( pInfo->db_getname(), pName ) == 0 ) {
	  store_unloadplug();
	  return pluginname;
	}

	store_unloadplug();

      }

    }

  }

  closedir( pDir );
  free( pluginname );
  return NULL;

}

int store_connect( char *pConnUrl ) {

  char *pPlugname = NULL;
  char *pProvider;

  pProvider = getparamvalue( pConnUrl, "provider" );

  if ( !pProvider ) {
    fprintf(stderr, "No provider (provider=) specified, use -S to view available providers\n");
    return 1;
  }
  
  if ( ( pPlugname = store_getplugbyname( pProvider ) ) ) {

    if ( !store_loadplug(pPlugname) ) {
      free( pPlugname );
      freeparamvalue( pProvider );
      fprintf(stderr, "ERROR: store_loadplug() failed\n");
      return 1;
    }
    else
      free( pPlugname );
      freeparamvalue( pProvider );

  }
  else {
     freeparamvalue( pProvider );
     return STORE_PROVIDER_RESOLVE_FAILED;
  }

  if ( pInfo ) {
    pConn = pInfo->db_connect( pConnUrl );

    if ( !pConn )
      return 1;
  }

  return 0;

}

void store_disconnect( ) {
  if ( pConn && pInfo )
    pInfo->db_disconnect( pConn );

  store_unloadplug();

}

int store_createchksum( struct host *pHost ) {

  char txt[8192];
  unsigned char *pMD5;
  int nSize = 0;
  struct property *pProps;

  txt[0] = 0;
  nSize = sizeof( pHost->mac ) + sizeof( pHost->addr ) + sizeof( pHost->name );
  nSize += sizeof( pHost->type );

  strcpy( txt, pHost->mac );
  strcat( txt, pHost->addr );
  strcat( txt, pHost->name );
  strcat( txt, pHost->type );

  pProps = pHost->prop;

  while ( pProps ) {
    nSize += strlen( pProps->val );
    
    if ( nSize > sizeof( txt ) -1 )
      return -1;

    strcat( txt, pProps->val );

    pProps = pProps->next;
  }
  
  txt[nSize] = 0;

  pMD5 = createMD5( txt, nSize );
  md5ToAscii( (char *)pHost->chksum, pMD5 );
  pHost->chksum[sizeof(pHost->chksum)] = 0;
  free( pMD5 );

  /*
  fprintf( stderr, "TXT=%s\n", txt );
  fprintf( stderr, "MD5=%s\n", pHost->chksum );
  */

  return 0;

}

int store_storehost(const struct host *pHost) {

  struct host *loop = (struct host *)pHost;
  int nErr;

  while ( loop ) {

    if ( store_createchksum( loop ) != 0 )
      return STORE_CREATECHECKSUM_FAILED;

    if ( ( nErr = pInfo->db_storehost( pConn, loop->mac, loop->addr, 
				       loop->name, loop->type, 
				       loop->chksum ) ) != 0 ) {

      if ( nErr == STORE_STOREHOST_DUPLICATE ) { /*dup*/
	return STORE_STOREHOST_DUPLICATE;
      }
      else {
	return STORE_STOREHOST_FAILED;
      }
    
    }

    if ( ( nErr = store_storeproperty( loop->prop, loop->chksum ) ) != 0 )
      return STORE_STOREPROP_FAILED;

    loop = loop->next;
  }

  return 0;

}


int store_storeproperty(const struct property *pProp, char *chksum ) {

  struct property *loop = (struct property *)pProp;
  int nErr = 0;

  while ( loop ) {

    if ( ( nErr = pInfo->db_storeproperty( pConn, loop->prop,
					   loop->val, chksum ) ) != 0 )
      return nErr;

    loop = loop->next;
  }

  return nErr;

}

