/* 
   Passifist
   Copyright (C) Patrik Karlsson <patrik@cqure.net> 2003
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "openssl/md5.h"


unsigned char *createMD5( const char *pData, const int nLen ) {

  MD5_CTX ctx;
  unsigned char *md5dig = ( unsigned char * ) malloc ( 17 );

  if ( !md5dig )
    return NULL;

  MD5_Init( &ctx );
  MD5_Update( &ctx, pData, nLen );
  MD5_Final( md5dig, &ctx );
  md5dig[16] = '\0';
  return md5dig;

}

void md5ToAscii( char *md5ascii, const unsigned char *md5 ) {

  int i;
  unsigned char hex[3];
  char ascii[40];

  memset( ascii, 0, sizeof( ascii ) );

  for ( i=0; i<16; i++ ) {
    memset( hex, 0, 3 );
    sprintf( (char *)&hex, "%.2X", md5[i] );
    strncat( ascii, hex, 2 );
  }

  memcpy( md5ascii, ascii, 32 );

} 


/*
 * Connection URL parsing code
 *
 */
char *getparamvalue( const char *pConnUrl, const char *pParam ) {

  char *pVal, *pRetVal, *pNewParam;
  int nSize = 255;
  int nStart = 0, nEnd = 0;

  if ( strlen( pParam ) > nSize )
    return NULL;

  pNewParam = (char *) malloc(strlen( pParam ) + 2);
  bzero( pNewParam, strlen( pParam ) + 2 );
  strncpy( pNewParam, pParam, strlen( pParam ) );
  strncat( pNewParam, "=", 1);

  pVal = strstr( pConnUrl, pNewParam );

  if ( !pVal )
    goto error;

  nStart = strlen( pParam ) + 1;

  /* search for '=' */
  if ( strlen( pVal ) < nStart )
    goto error;

  nEnd = nStart;

  while ( pVal[nEnd] != 0 && pVal[nEnd] != ';' )
    nEnd ++;

  if ( ( nEnd - nStart ) > nSize )
    goto error;

  pRetVal = (char *) malloc( nEnd - nStart + 1);
  strncpy( pRetVal, pVal + nStart, ( nEnd - nStart ) );
  pRetVal[nEnd - nStart] = 0;

  free( pNewParam );
  return pRetVal;

 error:
  free( pNewParam );
  return NULL;

}

void freeparamvalue( char *pVal ) {
  if ( pVal )
    free( pVal );
}
